#!/usr/bin/env python3
from orger import Mirror
from orger.inorganic import node, link, Quoted
from orger.common import dt_heading, error

from typing import Optional, List, Iterator, Any
from datetime import datetime

from more_itertools import bucket

from my.core.common import asdict
from pprint import pformat
def pp_item(i, **kwargs) -> str:
    # annoying, pprint doesn't have dataclass support till 3.10 https://bugs.python.org/issue43080
    return pformat(asdict(i), **kwargs)


class Auto(Mirror):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # TODO move this functionality to base module? might be useful for all modules
        self.extra_warnings: List[str] = []

        # these will be set below
        self.group_by_attr: Optional[str] = None
        self.body_attr    : Optional[str] = None
        self.title_attr   : Optional[str] = None


    def get_items(self) -> Mirror.Results:
        args = self.cmdline_args; assert args is not None
        self.group_by_attr = args.group_by
        self.body_attr     = args.body
        self.title_attr    = args.title
        return self.auto(it=args.name)


    # TODO hmm maybe reuse HPI core.query.select?
    # could abuse that for grouping by, sorting, raising errors, filtering fields, etc
    def auto(self, it) -> Mirror.Results:
        group_by = self.group_by_attr
        if isinstance(it, str):
            # treat as HPI query target
            from my.core.__main__ import _locate_functions_or_prompt
            [it] = _locate_functions_or_prompt([it], prompt=False)

        if group_by is None:
            for i in it():
                if isinstance(i, Exception):
                    yield error(i)
                else:
                    yield from self.render_one(i)
        else:
            good = []
            for i in it():
                if isinstance(i, Exception):
                    yield error(i)
                    continue
                good.append(i)

            _group_by = group_by  # make mypy happy
            groups = bucket(good, key=lambda i: getattr(i, _group_by))
            for key in groups:
                group = list(groups[key])

                def chit():
                    for i in group:
                        yield from self.render_one(i)
                children = list(chit())

                yield node(
                    heading=f'{group_by}: {key}',
                    children=children,
                )
        if len(self.extra_warnings) > 0:
            self.file_header = self.file_header + '\n'.join('# ' + w for w in ['', *self.extra_warnings]) + '\n'


    def _warn(self, w: str):
        if w not in self.extra_warnings:
            self.extra_warnings.append(w)


    def render_one(self, i) -> Iterator[node]:
        self._warn('WARNING: Default renderer is used! Implement render_one if you want nicer rendering')

        d = asdict(i)
        cls = i.__class__
        del i # delete from scope to avoid using by accident

        datetimes = [(k, v) for k, v in d.items() if isinstance(v, datetime)]
        dt: Optional[datetime] = None
        if len(datetimes) == 1:
            [(k, dt)] = datetimes
            # todo maybe warn that datetime is guessed
            del d[k] # probs no need to press twice?
            self._warn(f"NOTE: {cls} is using '{k}' as the timestamp")
        else:
            self._warn(f"WARNING: {cls} couldn't guess timestamp: expected single datetime field, got {datetimes}")

        # todo could pass fmt string for group as well?
        if self.group_by_attr is not None:
            try:
                d.pop(self.group_by_attr)
            except KeyError as e:
                yield error(e)

        def fmt_attr(attr: Optional[str]) -> Optional[str]:
            if attr is None:
                return None

            import string
            fake_self: Any = object()  # fine to call it as class method here..
            fields = [
                f
                for (_, f, _, _) in string.Formatter.parse(fake_self, attr)
                if f is not None # might be none for the last bit (before the static bit)
            ]
            fstr: str
            if len(fields) == 0:
                # must be field name without formatting string?
                fields = [attr]
                fstr = '{' + attr + '}'
            else:
                fstr = attr
            res = fstr.format(**d)
            for f in fields:
                # should succeed after format call?
                d.pop(f)
            return res


        title = fmt_attr(self.title_attr)
        body  = fmt_attr(self.body_attr)

        node_title = cls.__name__ if title is None else title
        node_body = Quoted(pp_item(d, width=120)).quoted()
        if body is not None:
            node_body += '\n' + body

        yield node(
            heading=dt_heading(dt, node_title),
            body=node_body,
        )


def setup_parser(p) -> None:
    p.add_argument(
        '--name',
        type=str,
        required=True,
        help="HPI function name (see 'hpi query --help')",
    )
    p.add_argument('--group-by', type=str, default=None, help='field to group items by')
    p.add_argument('--body'    , type=str, default=None, help='field to use as item body')
    p.add_argument('--title'   , type=str, default=None, help='field to use as item title')
    # will support later
    # p.add_argument('--id'      , type=str, default=None, help='field to use as item ID')


if __name__ == '__main__':
    Auto.main(setup_parser=setup_parser)



#!/usr/bin/env python3

from orger import Mirror
from orger.inorganic import node, link, docview_link, literal
from orger.common import dt_heading, error

from typing import Optional
from textwrap import indent, wrap

from more_itertools import bucket


class config:
    MAX_LINE_WIDTH: int = 120


colors_doc = """
To highlight tags based on color, add this to your Emacs config:
  (with-eval-after-load 'org
    (dolist (color '("yellow" "red" "purple" "blue" "green"))
        (add-to-list 'org-tag-faces
                     `(,color . (:background ,color)))))

NOTE: you'll need to restart emacs for this to take effect.
If some colors look too bright for you, you can just manually map them instead, e.g.:
  (with-eval-after-load 'org
    (add-to-list 'org-tag-faces '("red"   . (:foreground "#cc0000")))
    (add-to-list 'org-tag-faces '("green" . (:foreground "#00aa00"))))
"""


class Zotero(Mirror):
    DEFAULT_HEADER = Mirror.DEFAULT_HEADER + indent(colors_doc, '# ', lambda line: True)

    def get_items(self) -> Mirror.Results:
        from my import zotero

        errors = []
        good = []
        for a in zotero.annotations():
            if isinstance(a, Exception):
                errors.append(a)
            else:
                good.append(a)

        for e in errors:
            yield error(e)


        groups = bucket(good, key=lambda a: a.item)
        for item in groups:
            file_annotations = groups[item]
            def chit():
                for a in file_annotations:
                    parts = []
                    text = a.text
                    if text is not None:
                        # todo not sure about it here... maybe should rely on softwrap in emacs instead?
                        text = '\n'.join(wrap(text, width=config.MAX_LINE_WIDTH))
                        text = literal(text)
                        parts.append(text)
                    comment = a.comment
                    if comment is not None:
                        parts.append(comment)
                    page1 = a.page + 1  # NOTE: zotero using 0-indexing, pdfview using 1-indexing
                    body = '\n'.join(parts)

                    color = a.color_human
                    tags = list(a.tags)
                    # todo not sure which is best?
                    tags.append(color)
                    properties = {
                        'ZOTERO_COLOR': color,
                    }
                    if len(a.tags) > 0:
                        # zotero tags cal be multi-word? guess worth adding just in case
                        properties['ZOTERO_TAGS'] = ', '.join(a.tags)  # not sure what's the best separator...

                    # todo not sure about it...
                    mtodo: Optional[str] = None
                    if 'todo' in {t.lower() for t in tags}:
                        mtodo = 'TODO'

                    heading = docview_link(path=item.file, title=f'page {page1}', page1=page1)
                    if comment is not None:
                        # try to display first few words?
                        cline = wrap(comment, width=config.MAX_LINE_WIDTH)[0]
                        heading = heading + ' ' + cline
                    # todo would be nice to align tags, maybe...
                    yield node(
                        todo=mtodo,
                        heading=dt_heading(
                            a.added,
                            heading,
                        ),
                        tags=tags,
                        properties=properties,
                        body=body,
                    )
            body = ''
            url = item.url
            if url is not None:
                body = url
            yield node(
                heading=docview_link(path=item.file, title=item.title),
                tags=item.tags,
                body=body,
                children=list(chit())
            )

if __name__ == '__main__':
    Zotero.main()


# TODO maybe handle
# (see zotero-protocol-handler.js)
# zotero://pdf.js/viewer.html
# zotero://pdf.js/pdf/1/ABCD5678



#!/usr/bin/env python3
from orger import Mirror
from orger.inorganic import node, link, Quoted
from orger.common import dt_heading, error

from my.vk.favorites import favorites, Favorite

class VkFavs(Mirror):
    def get_items(self) -> Mirror.Results:
        for f in favorites():
            if isinstance(f, Favorite):
                yield node(
                    heading=dt_heading(
                        f.dt,
                        f.title if f.url is None else link(url=f.url, title=f.title),
                    ),
                    body=Quoted(f.text),
                )
            else: # Exception
                yield error(f)


if __name__ == '__main__':
    VkFavs.main()



#!/usr/bin/env python3

from orger import Mirror
from orger.inorganic import node, link, Quoted
from orger.common import dt_heading, error
from orger import pandoc

from more_itertools import bucket

from my.zulip.organization import messages


class Zulip(Mirror):
    def get_items(self) -> Mirror.Results:
        good = []
        for m in messages():
            if isinstance(m, Exception):
                yield error(m)
                continue
            good.append(m)

        groups = bucket(good, key=lambda m: m.subject)
        for subject in groups:
            group = list(groups[subject])

            def chit():
                for m in group:
                    yield node(
                        heading=dt_heading(m.sent, link(title=m.sender.full_name, url=m.permalink)),
                        body=pandoc.to_org(m.content, from_='markdown'),
                    )

            yield node(
                heading=dt_heading(group[0].sent, subject), # TODO link
                children=list(chit()),
            )


test = Zulip.make_test(
    heading='Tracking Books',
    contains='progress from my Kobo to Goodreads',
)


if __name__ == '__main__':
    Zulip.main()



#!/usr/bin/env python3
from orger import Mirror
from orger.inorganic import node, link, Quoted
from orger.common import dt_heading

import my.stackexchange.stexport as se


class Stackexchange(Mirror):
    def get_items(self) -> Mirror.Results:
        # TODO adapt for other stackexchange items
        se_data = se.site('stackoverflow')
        for q in se_data.questions:
            # TODO could emit items along with level, then it would look a bit more natural
            yield node(
                dt_heading(q.creation_date, link(url=q.link, title=q.title)),
                tags=q.tags,
                body=Quoted(q.body_markdown),
                # todo eh, would be useful to have md2org perhaps?
            )


if __name__ == '__main__':
    Stackexchange.main()


# todo how to simplify this even further? i.e. so you
# - don't need the class (just a function?)
#   maybe yield the marker object or something (policy?)
# - don't have to write main()
# - maybe not even do imports?
